/*
 *   To compile:
 *
 *   gcc -fPIC -shared -o nvidia_spank.so nvidia_spank.c -lnvidia-ml -L/usr/local/cuda-8.0/targets/x86_64-linux/lib/stubs
 *
 */
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/resource.h>

#include <slurm/spank.h>

/*
 *  * All spank plugins must define this macro for the
 *   * Slurm plugin loader.
 *    */
SPANK_PLUGIN(nvidia_spank, 1);

static int nvidia_compute_found = 0;

static int nvidia_accounting = 0;
static int nvidia_accounting_found = 0;

static int nvidia_mps = 0;
static int nvidia_mps_found = 0;

static int _nvidia_compute_opt_process (int val,
                                const char *optarg,
                                int remote);
static int _nvidia_accounting_opt_process (int val,
                                const char *optarg,
                                int remote);

static int _nvidia_mps_opt_process (int val,
                                const char *optarg,
                                int remote);

struct spank_option gpu_compute_opt =
    { "gpu-compute", "[default|exclusive-thread|exclusive-process|exclusive]",
      "Select NVidia GPU compute mode, it can be set to default, exclusive-thread and exclusive-process (same as exclusive)", 1, 2,
      (spank_opt_cb_f) _nvidia_compute_opt_process
    };

struct spank_option gpu_accounting_opt =
    { "gpu-accounting", NULL,
      "Turn on NVidia GPU accounting", 0, 1,
      (spank_opt_cb_f) _nvidia_accounting_opt_process
    };

struct spank_option gpu_mps_opt =
    { "gpu-mps", NULL,
      "Start MPS server, it will automaticall switch to exclusive thread mode", 0, 1,
      (spank_opt_cb_f) _nvidia_mps_opt_process
    };

int slurm_spank_init (spank_t sp, int ac, char **av)
{
    int sc = spank_context();
    if (sc == S_CTX_ALLOCATOR || sc == S_CTX_REMOTE || sc == S_CTX_LOCAL) {
      spank_option_register (sp, &gpu_compute_opt);
      spank_option_register (sp, &gpu_accounting_opt);
      spank_option_register (sp, &gpu_mps_opt);
    }
    return (0);
}

int slurm_spank_task_init(spank_t sp, int ac, char **av)
{
    char val[1024]={0};

    if (nvidia_mps) {
        // Try to check if running, assume failure means that it is not
        if(WEXITSTATUS(system("echo get_server_list | /bin/nvidia-cuda-mps-control >/dev/null 2>&1"))==1) {
            slurm_info("Starting MPS");
            setenv("CUDA_DEVICE_MAX_CONNECTIONS","8",0);
            setenv("CUDA_MPS_LOG_DIRECTORY","/tmp",0);
            system("/bin/nvidia-cuda-mps-control -d");
        }
    }
}

static int _nvidia_compute_opt_process (int val,
                                const char *optarg,
                                int remote) {
    if (nvidia_mps) {
        slurm_error("MPS option is on, ignoring compute option %s",optarg);
    } else if(optarg) {
        nvidia_compute_found = 1;
    }
    return (0);
}

static int _nvidia_accounting_opt_process (int val,
                                const char *optarg,
                                int remote) {
    nvidia_accounting = 1;
    nvidia_accounting_found = 1;
    return (0);
}

static int _nvidia_mps_opt_process (int val,
                                const char *optarg,
                                int remote) {
    nvidia_mps = 1;
    nvidia_mps_found = 1;
    // also force exclusive mode for compute
    nvidia_compute_found = 1;
    return (0);
}
